@@ -29,6 +29,7 @@
 ;;    b) if fixed/float use `int *x;' etc
 ;;    c> if function use `void *x;'
 ;; 6) Check use of comments as attributes.
+;; 7) maybe move to munge-base: typedef-decl? udict->typedef-names
 
 ;;; Code:
 
@@ -43,7 +44,7 @@
 	    ;; munging
 	    stripdown-udecl
 	    udecl-rem-type-qual specl-rem-type-qual
-	    udecl->mdecl udecl->mdecl/comm mdecl->udecl
+	    udecl->mdecl/comm mdecl->udecl
 
 	    unwrap-decl
 	    canize-enum-def-list
@@ -54,17 +55,20 @@
 	    unitize-decl unitize-comp-decl unitize-param-decl
 	    declr-ident declr-id decl-id
 	    iter-declrs
-	    split-decl 
+	    split-decl
 
+	    udict->typedef-names
 	    inc-keeper?
 
 	    ;; debugging
 	    stripdown-1
 	    tdef-splice-specl
 	    tdef-splice-declr)
-  #:re-export (expand-typerefs clean-field-list clean-fields split-udecl)
-  #:use-module ((nyacc lang c99 cpp) #:select (eval-cpp-expr))
-  #:use-module (nyacc lang c99 cxeval) ;; eval-c99-cx
+  #:re-export (expand-typerefs
+	       reify-declr reify-decl
+	       udecl->mdecl split-udecl
+	       clean-field-list)
+  #:use-module (nyacc lang c99 cxeval)
   #:use-module (nyacc lang c99 munge-base)
   #:use-module (nyacc lang c99 pprint)
   #:use-module (nyacc lang c99 util)
@@ -118,9 +122,9 @@
 ;; @end example
 
 ;; mdecl is
-;; ("foo" (pointer-to) (array-of 3) (fixed-type "unsigned int"))
+;; ("foo" (pointer-to) (ary-declr 3) (fixed-type "unsigned int"))
 ;; which can be converted to
-;; ("(*foo) (array-of 3) (fixed-type "unsigned int"))
+;; ("(*foo) (ary-declr 3) (fixed-type "unsigned int"))
 ;; which can be converted to
 ;; (("((*foo)[0])" (fixed-type "unsigned int"))
 ;;  ("((*foo)[1])" (fixed-type "unsigned int"))
@@ -134,16 +138,13 @@
 ;; it should be processed.
 ;; @end deffn
 (define (inc-keeper? tree filter)
-  (if (and (eqv? (sx-tag tree) 'cpp-stmt)
-	   (eqv? (sx-tag (sx-ref tree 1)) 'include)
-	   (pair? (sx-ref* tree 1 2))
-	   (if (procedure? filter)
-	       (let ((incl (sx-ref* tree 1 1))
-		     (path (sx-attr-ref (sx-ref tree 1) 'path)))
-		 (filter incl path))
-	       filter))
-      (sx-ref* tree 1 2)
-      #f))
+  (sx-match tree
+    ((cpp-stmt (include (@ (path ,path)) ,spec ,tree))
+     (and (if (procedure? filter) (filter spec path) filter) tree))
+    ((cpp-stmt (include-next (@ (path ,path)) ,spec ,tree))
+     (and (if (procedure? filter) (filter spec path) filter) tree))
+    ((cpp-stmt . ,rest) #f)
+    (,_ #f)))
 
 ;; @deffn {Procedure} c99-trans-unit->udict tree [seed] [#:inc-filter f]
 ;; @deffnx {Procedure} c99-trans-unit->udict/deep tree [seed]
@@ -194,6 +195,7 @@
        seed
        (cdr tree))
       seed))
+
 (define (c99-trans-unit->udict/deep tree)
   (c99-trans-unit->udict tree #:inc-filter #t))
 
@@ -300,7 +302,6 @@
     (let*-values (((tag decl-attr specl declrs) (split-decl decl))
 		  ((tag) (values 'udecl)))
       ;; TODO: for typedefs add attr (typedef "name") to associated udecls
-      ;;(sferr "\nsx-match specl: attr=~S\n" attr) (pperr specl) (pperr declrs)
       (sx-match specl
 
 	;; struct typedefs 
@@ -454,8 +455,8 @@
     ((init-declr ,declr . ,rest) (declr-ident declr))
     ((comp-declr ,declr) (declr-ident declr))
     ((param-declr ,declr) (declr-ident declr))
-    ((array-of ,dir-declr ,array-spec) (declr-ident dir-declr))
-    ((array-of ,dir-declr) (declr-ident dir-declr))
+    ((ary-declr ,dir-declr ,array-spec) (declr-ident dir-declr))
+    ((ary-declr ,dir-declr) (declr-ident dir-declr))
     ((ptr-declr ,pointer ,dir-declr) (declr-ident dir-declr))
     ((ftn-declr ,dir-declr . ,rest) (declr-ident dir-declr))
     ((scope ,declr) (declr-ident declr))
@@ -521,17 +522,25 @@
   '("int8_t" "uint8_t" "int16_t" "uint16_t"
     "int32_t" "uint32_t" "int64_t" "uint64_t"))
 
+;; === enums and defines ===============
+
 ;; @deffn {Procedure} typedef-decl? decl)
 ;; @end deffn
 (define (typedef-decl? decl)
-  (sx-match decl
-    ((decl (decl-spec-list (stor-spec (typedef)) . ,r1) . ,r2) #t)
-    (,_ #f)))
+  (eq? 'typedef (and=> (sx-ref* decl 1 1 1) sx-tag)))
+
+(define* (udict->typedef-names udict #:optional (names '()))
+  (fold (lambda (pair names)
+	  (let ((name (car pair)) (decl (cdr pair)))
+	    (if (typedef-decl? decl) (cons name names) names)))
+	names udict))
 
 ;; === enums and defines ===============
 
-;; @deffn {Procedure} c99-trans-unit->ddict tree [seed] [#:inc-filter proc]
+;; @deffn {Procedure} c99-trans-unit->ddict tree [seed] [#:key ...]
 ;; Extract the #defines from a tree as
+;; @*
+;; Example:
 ;; @example
 ;;  (define (name "ABC") (repl "repl"))
 ;;  (define (name "MAX") (args "X" "Y") (repl "(X)..."))
@@ -546,31 +555,26 @@
 (define* (c99-trans-unit->ddict tree
 				#:optional (ddict '())
 				#:key inc-filter skip-fdefs)
-  (define (can-def-stmt defn)
-    (let* ((tail (sx-tail defn 1))
-	   (name (car (assq-ref tail 'name)))
-	   (args (assq-ref tail 'args))
-	   (repl (car (assq-ref tail 'repl))))
-      (cons name (if args (cons args repl) repl))))
-  (define (cpp-def? tree)
-    (if (and (eq? 'cpp-stmt (sx-tag tree))
-	     (eq? 'define (sx-tag (sx-ref tree 1)))
-	     (not (and skip-fdefs (assq 'args (sx-tail (sx-ref tree 1) 1)))))
-	(can-def-stmt (sx-ref tree 1))
-	#f))
+  (define (can-def-stmt tree)
+    (sx-match tree
+      ((cpp-stmt (define (name ,name) (repl ,repl)))
+       (cons name repl))
+      ((cpp-stmt (define (name ,name) (args . ,args) (repl ,repl)))
+       (if skip-fdefs #f (cons name (cons args repl))))
+      (,_ #f)))
+
   (if (pair? tree)
       (fold
        (lambda (tree ddict)
 	 (cond
-	  ((cpp-def? tree) =>
-	   (lambda (def-stmt)
-	     (cons def-stmt ddict)))
+	  ((can-def-stmt tree) => (lambda (def) (cons def ddict)))
 	  ((inc-keeper? tree inc-filter) =>
 	   (lambda (tree)
-	     (c99-trans-unit->ddict tree ddict #:inc-filter inc-filter)))
+	     (c99-trans-unit->ddict tree ddict
+				    #:inc-filter inc-filter
+				    #:skip-fdefs skip-fdefs)))
 	  (else ddict)))
-       ddict
-       (cdr tree))
+       ddict (sx-tail tree))
       ddict))
 
 ;; @deffn {Procedure} udict-enums->ddict udict [ddict] => defs
@@ -586,23 +590,21 @@
 ;; @end example
 ;; @end deffn
 (define* (udict-enums->ddict udict #:optional (ddict '()))
-  (define (gen-nvl enum-def-list ddict)
-    (let ((def-list (and=> (canize-enum-def-list enum-def-list udict ddict)
-			   cdr)))
-    (fold (lambda (edef ddict) ;; (enum-def (ident ,name) (fixed ,val) ...
-	    (acons (sx-ref* edef 1 1) (sx-ref* edef 2 1) ddict))
-	  ddict def-list)))
+  (define (gen-nvl edef-list ddict)
+    (let ((def-list (and=> (canize-enum-def-list edef-list udict ddict) cdr)))
+      (fold (lambda (edef dd) (acons (sx-ref* edef 1 1) (sx-ref* edef 2 1) dd))
+	    ddict def-list)))
   (fold
    (lambda (pair ddict)
-     (if (and (pair? (car pair)) (eq? 'enum (caar pair)))
-	 ;; (enum . ,name) ...
-	 (let* ((specl (sx-ref (cdr pair) 1))
-		(tspec (car (assq-ref specl 'type-spec))))
-	   (if (not (eq? 'enum-def (car tspec)))
-	       (throw 'nyacc-error "udict-enums->ddict: expecting enum-def"))
-	   (gen-nvl (assq 'enum-def-list (cdr tspec)) ddict))
-	 ;; else ...
-	 ddict))
+     (let* ((specs (sx-ref (cdr pair) 1))
+	    (tspec (sx-ref (sx-find 'type-spec specs) 1)))
+       (sx-match tspec
+	 ((enum-def (ident ,name) (enum-def-list . ,defs))
+	  (gen-nvl (sx-ref tspec 2) ddict))
+	 ((enum-def (enum-def-list . ,defs))
+	  (gen-nvl (sx-ref tspec 1) ddict))
+	 (,_
+	  ddict))))
    ddict udict))
 
 
@@ -621,10 +623,12 @@
 ;; @end example
 ;; @noindent
 ;; @end deffn
-(define* (canize-enum-def-list enum-def-list #:optional (udict '()) (ddict '()))
-  (define (fail ident)
-    (sferr "*** munge: failed to convert enum ~S to constant\n" (sx-ref ident 1))
-    #f)
+(define* (canize-enum-def-list enum-def-list
+			       #:optional (udict '()) (ddict '())
+			       #:key fail-proc)
+  (define (fail fmt . args)
+    (and fail-proc (apply fail-proc fmt args)))
+
   (let loop ((rez '()) (nxt 0) (ddict ddict) (edl (sx-tail enum-def-list 1)))
     (cond
      ((null? edl)
@@ -634,21 +638,17 @@
 	((enum-defn (@ . ,attr) ,ident)
 	 (let ((sval (number->string nxt)))
 	   (loop (cons (sx-list 'enum-defn attr ident `(fixed ,sval)) rez)
-		 (1+ nxt)  (acons (sx-ref ident 1) sval ddict) (cdr edl))))
-	((enum-defn ,ident)
-	 (let ((sval (number->string nxt)))
-	   (loop (cons (sx-list 'enum-defn #f ident `(fixed ,sval)) rez)
-		 (1+ nxt)  (acons (sx-ref ident 1) sval ddict) (cdr edl))))
-	((enum-defn (@ . attr) ,ident ,expr)
-	 (let* ((ival (or (eval-c99-cx expr udict ddict) (fail ident) nxt))
-		(sval (number->string iv)))
-	   (loop (cons (sx-list 'enum-defn attr ident `(fixed ,sval)) rez)
-		 (1+ ival) (acons (sx-ref ident 1) sval ddict) (cdr edl))))
-	((enum-defn ,ident ,expr)
-	 (let* ((ival (or (eval-c99-cx expr udict ddict) (fail ident) nxt))
-		(sval (number->string ival)))
-	   (loop (cons (sx-list 'enum-defn #f ident `(fixed ,sval)) rez)
-		 (1+ ival) (acons (sx-ref ident 1) sval ddict) (cdr edl)))))))))
+		 (1+ nxt) (acons (sx-ref ident 1) sval ddict) (cdr edl))))
+	((enum-defn (@ . ,attr) ,ident ,expr)
+	 (let* ((ival (eval-c99-cx expr udict ddict))
+		(sval (and (number? ival) (number->string ival))))
+	   (cond
+	    (sval
+	     (loop (cons (sx-list 'enum-defn attr ident `(fixed ,sval)) rez)
+		   (1+ ival) (acons (sx-ref ident 1) sval ddict) (cdr edl)))
+	    (else
+	     (fail "munge/canize-enum-def-list: can't expand ~S"(sx-ref ident 1))
+	     (loop rez nxt ddict (cdr edl)))))))))))
 
 ;; @deffn {Procecure} enum-ref enum-def-list name => string
 ;; Gets value of enum where @var{enum-def-list} looks like
@@ -763,7 +763,6 @@
 ;; Remove cruft from declaration-specifiers (tail). ??
 ;; @end deffn
 (define* (strip-decl-spec-tail dsl-tail #:key keep-const?)
-  ;;(simple-format #t "spec=tail: ~S\n" dsl-tail)
   (let loop ((dsl1 '()) (const-seen? #f) (tail dsl-tail))
     (if (null? tail)
 	(reverse (if (and const-seen? keep-const?)
@@ -781,135 +780,6 @@
 
 ;; === munged specification ============
 
-(define (def-namer) (symbol->string (gensym "@")))
-
-;; @deffn {Procedure} udecl->mdecl udecl [#:namer def-namer]
-;; @deffnx {Procedure} udecl->mdecl/comm udecl [#:def-comm ""]
-;; Turn a stripped-down unit-declaration into an m-spec.  The second version
-;; includes the comment. This assumes decls have been run through
-;; @code{stripdown}.
-;; @example
-;; (decl (decl-spec-list (type-spec "double"))
-;;       (init-declr-list (
-;;       (comment "state vector")
-;; =>
-;; ("x" "state vector" (array-of 10) (float "double")
-;; @end example
-;; @noindent
-;; The optional keyword argument @var{namer} is a procdedure returning a string
-;; to add for abstract declarators.  If an identifier is not provided, a
-;; random identifier starting with @code{@} will be provided.
-;; @end deffn
-(define* (udecl->mdecl decl #:key (namer def-namer))
-
-  ;; Hmm.  We convert array size back to C code (string).  Now that I am working
-  ;; on constant expression eval (eval-c99-cx) maybe we should change that.
-  (define (cnvt-size-expr size-spec)
-    ;;(with-output-to-string (lambda () (pretty-print-c99 size-spec)))
-    size-spec)
-
-  (define (unwrap-specl specl)
-    (and=> (assq-ref (sx-tail specl) 'type-spec) car))
-
-  (define (unwrap-pointer pointer)  ;; =>list IGNORES TYPE QUALIFIERS
-    ;;(sferr "unwrap-pointer ~S\n" pointer)
-    (sx-match pointer
-      ((pointer (type-qual-list . ,type-qual) ,pointer)
-       (cons '(pointer-to) (unwrap-pointer pointer)))
-      ((pointer (type-qual-list . ,type-qual)) '((pointer-to)))
-      ((pointer ,pointer) (cons '(pointer-to) (unwrap-pointer pointer)))
-      ((pointer) '((pointer-to)))
-      (,_
-       (sferr "unwrap-pointer failed on:\n") (pperr pointer)
-       (throw 'nyacc-error "unwrap-pointer"))))
-
-  (define (make-abs-dummy) ;; for abstract declarator make a dummy
-    (namer))
-  (define (make-abs-dummy-tail)
-    (list (make-abs-dummy)))
-  
-  (define* (unwrap-declr declr #:key (const #f))
-    ;;(sferr "unwrap-declr:\n") (pperr declr #:per-line-prefix "  ")
-    (sx-match declr
-      ((ident ,name)
-       (list name))
-      ((init-declr ,item)
-       (unwrap-declr item #:const const))
-
-      ((array-of ,dir-declr ,size)
-       (cons `(array-of ,(cnvt-size-expr size)) (unwrap-declr dir-declr)))
-      ((array-of ,dir-declr)
-       (cons `(array-of) (unwrap-declr dir-declr)))
-
-      ((ftn-declr ,dir-declr ,param-list)
-       (cons `(function-returning ,param-list) (unwrap-declr dir-declr)))
-      ((abs-ftn-declr ,dir-abs-declr)
-       (cons `(function-returning) (unwrap-declr dir-abs-declr)))
-      ((abs-ftn-declr ,dir-abs-declr ,param-list)
-       (cons `(function-returning ,param-list) (unwrap-declr dir-abs-declr)))
-      ;;((anon-ftn-declr ,param-list) ???
-
-      ((scope ,expr) (unwrap-declr expr))
-
-      ((ptr-declr ,pointer ,dir-declr)
-       (let ((res (append (unwrap-pointer pointer) (unwrap-declr dir-declr))))
-	 (if const (cons '(const) res) res)))
-
-      ;; abstract declarator and direct abstract declarator
-      ((abs-declr ,pointer ,dir-abs-declr)
-       (append (unwrap-pointer pointer) (unwrap-declr dir-abs-declr)))
-      ((abs-declr (pointer))
-       (append (unwrap-pointer (sx-ref declr 1)) (make-abs-dummy-tail)))
-      ((abs-declr (pointer ,pointer-val))
-       (cons* '(pointer-to) (make-abs-dummy-tail)))
-      ((abs-declr ,dir-abs-declr)
-       (unwrap-declr dir-abs-declr))
-
-      ;; declr-scope
-      ;; declr-array dir-abs-declr
-      ;; declr-array dir-abs-declr assn-expr
-      ;; declr-array dir-abs-declr type-qual-list
-      ;; declr-array dir-abs-declr type-qual-list assn-expr
-      ((declr-scope ,abs-declr)		; ( abs-declr )
-       (unwrap-declr abs-declr))
-      ((declr-array ,dir-abs-declr)	; []
-       (cons '(array-of "") (make-abs-dummy-tail))) ;; ???
-      ((declr-array ,dir-abs-declr (type-qual-list . ,type-quals))
-       ;; TODO: deal with "const" type-qualifier
-       (cons '(array-of "") (make-abs-dummy-tail)))
-      ((declr-array ,dir-abs-declr ,assn-expr)
-       (cons `(array-of ,assn-expr) (unwrap-declr dir-abs-declr)))
-      ((declr-array ,dir-abs-declr ,type-qual-list ,assn-expr)
-       ;; TODO: deal with "const" type-qualifier
-       (cons `(array-of ,assn-expr) (unwrap-declr dir-abs-declr)))
-
-      ((bit-field (ident ,name) ,size)
-       (list `(bit-field ,(cnvt-size-expr size)) name))
-
-      ((comp-declr ,item) (unwrap-declr item))
-      ((param-declr ,item) (unwrap-declr item))
-
-      (,_
-       (sferr "munge/unwrap-declr missed:\n")
-       (pperr declr)
-       (throw 'nyacc-error "c99/munge: udecl->mdecl failed")
-       #f)))
-
-  ;;(sferr "decl:\n") (pperr decl)
-  (let-values (((tag attr specl declr) (split-udecl decl)))
-    (let* ((tspec (sx-ref specl 1))	; type-spec
-	   (const (and=> (sx-ref specl 2)	; const pointer ???
-			 (lambda (sx) (equal? (sx-ref sx 1) "const"))))
-	   (declr (or (sx-ref decl 2)	; param-decl -- e.g., f(int)
-		      `(ident ,(make-abs-dummy))))
-	   (m-specl (unwrap-specl specl))
-	   (m-declr (unwrap-declr declr #:const const))
-	   (m-decl (reverse (cons m-specl m-declr))))
-      ;;(sferr "decl:\n") (pperr decl)
-      ;;(sferr "declr:\n") (pperr declr)
-      ;;(sferr "r-mdecl: ~S\n" (cons m-specl m-declr))
-      m-decl)))
-
 (define* (udecl->mdecl/comm decl #:key (def-comm ""))
   (let* ((comm (or (and=> (assq 'comment (sx-attr decl)) cadr) def-comm))
 	 (spec (udecl->mdecl decl)))
@@ -934,11 +804,17 @@
       (((pointer-to) . ,rest)
        (doit `(ptr-declr (pointer) ,declr) rest))
       (((array-of ,size) . ,rest)
-       (doit `(array-of ,declr ,size) rest))
-      
+       (doit `(ary-declr ,declr ,size) rest))
+
+      (((struct-ref (ident ,name))) (make-udecl (car mdecl-tail) declr))
+      (((union-ref (ident ,name))) (make-udecl (car mdecl-tail) declr))
+     (((struct-def (ident ,name))) (make-udecl (car mdecl-tail) declr))
+      (((union-def (ident ,name))) (make-udecl (car mdecl-tail) declr))
+
       (,_
        (sferr "munge/mdecl->udecl missed:\n")
        (pperr mdecl-tail)
+       (error "failed")
        (throw 'nyacc-error "munge/mdecl->udecl failed")
        #f)))
 
@@ -946,6 +822,5 @@
 	(rest (cdr mdecl)))
     (doit `(ident ,name) rest)))
 
-;; === deprecated ====================
 
 ;; --- last line ---
