@@ -1,6 +1,6 @@
 ;;; scripts/compile-ffi.scm --- NYACC's command-line FFI compiler
 
-;; Copyright (C) 2017-2019 Matthew R. Wette
+;; Copyright (C) 2017-2020 Matthew R. Wette
 ;;
 ;; This program is free software; you can redistribute it and/or
 ;; modify it under the terms of the GNU Lesser General Public
@@ -18,7 +18,6 @@
 ;;; Author: Matt Wette <mwette@alumni.caltech.edu>
 
 ;;; Commentary:
-
 ;; Usage: compile-ffi [ARGS]
 ;; Type `compile-ffi --help' for help.
 
@@ -30,9 +29,9 @@
   #:use-module ((system base compile) #:select (compile-file))
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-37)
-  #:version (1 01 2))
+  #:version (1 02 0))
 
-(define *ffi-help-version* "1.01.2")
+(define *ffi-help-version* "1.02.0")
 
 (define %summary
   "Compile a ffi-file (C interface spec) to Scheme (or maybe .go).")
@@ -73,6 +72,7 @@ Generate a Guile Scheme file from the source FFI file FILE.
   -I  --inc-dir=DIR    add DIR to list of dir's to search for C headers
   -o, --output=OFILE   write output to OFILE
   -d, --debug=x,y      set debug flags: echo-decl, parse
+  -D, --list-deps      list dependencies and quit
   -X, --no-exec        don't generate .go file(s)
   -R, --no-recurse     don't do recursive compile on dep's
 
@@ -102,6 +102,9 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
    (option '(#\I "inc-dir") #t #f
 	   (lambda (opt name arg opts files)
 	     (values (acons/seed 'inc-dirs arg opts) files)))
+   (option '(#\D "list-deps") #f #f
+	   (lambda (opt name arg opts files)
+	     (values (acons 'list-deps #t (acons 'no-recurse #t opts)) files)))
    (option '(#\R "dont-recurse") #f #f
 	   (lambda (opt name arg opts files)
 	     (values (acons 'no-recurse #t opts) files)))
@@ -157,7 +160,7 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
      (let* ((base (string-join (map symbol->string fmod) "/"))
 	    (xffi (find-in-path (string-append base ".ffi")))
 	    (xscm (find-in-path (string-append base ".scm"))))
-       (unless xffi (fail "compiled dependent ~S not found" fmod))
+       (unless xffi (fail "dependent ~S not found" fmod))
        (cond
 	((not xscm) (cons xffi seed))
 	((more-recent? xffi xscm) (cons xffi seed))
@@ -177,7 +180,7 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
   (check-ffi-deps (quote path-list) (find-ffi-uses attr ...)))
 (export define-ffi-module)
 
-(define (updated-ffi-deps file options)
+(define (outdated-ffi-deps file)
   (unless (access? file R_OK)
     (fail "not found: ~S" file))
   (call-with-input-file file
@@ -190,40 +193,60 @@ Report bugs to https://savannah.nongnu.org/projects/nyacc.\n"))
 	   ((and (pair? exp) (eqv? 'define-ffi-module (car exp))) (eval exp env))
 	   (else (loop (read iport)))))))))
 
+(define (gen-ffi-deps file)
+  (let loop ((odeps '()) (next '()) (prev (outdated-ffi-deps file)))
+    (cond
+     ((pair? prev)
+      (loop (cons (car prev) odeps)
+	    (if (member (car prev) odeps) next (cons (car prev) next))
+	    (cdr prev)))
+     ((pair? next)
+      (loop odeps '() next))
+     (else
+      (reverse
+       (fold
+	(lambda (odep deps) (if (member odep deps) deps (cons odep deps)))
+	'() odeps))))))
+
 (define (ensure-ffi-deps file options)
-  (for-each
-   (lambda (dep-file)
-     (cond
-      ((assq-ref options 'no-recurse)
-       (warn "please recompile `~A'" dep-file))
-      (else
-       (compile-ffi dep-file (acons 'no-exec #f options)))))
-   (updated-ffi-deps file options)))
+  (let ((options `((no-recurse . #t) (no-exec . #f) . ,options))
+	(ffi-deps (gen-ffi-deps file)))
+    (for-each
+     (lambda (dep-file) (compile-ffi dep-file options))
+     ffi-deps)))
 
 ;; -----------------------------------------------------------------------------
 
-(define (cleanup path)
-  (basename path))
+(define (fix-path path)
+  (let* ((cwd (getcwd)))
+    (if (string-contains path cwd)
+	(substring/shared path (1+ (string-length cwd)))
+	path)))
 
-(define* (compile-ffi ffi-file options #:key module)
+(define (compile-ffi ffi-file options)
   (let* ((base (string-drop-right ffi-file 4))
 	 (scm-file (string-append base ".scm")))
-    (ensure-ffi-deps ffi-file options)
+    (unless (assq-ref options 'no-recurse)
+      (ensure-ffi-deps ffi-file options))
+    (when (assq-ref options 'list-deps)
+      (for-each (lambda (f) (sfmt "~A\n" (fix-path f))) (gen-ffi-deps ffi-file))
+      (exit 0))
     (catch 'ffi-help-error
       (lambda ()
-	(sfmt "compiling `~A' ...\n" (cleanup ffi-file))
+	(sfmt "compiling `~A' ...\n" (fix-path ffi-file))
 	(compile-ffi-file ffi-file options)
-	(sfmt "... wrote `~A'\n" (cleanup scm-file)))
+	(sfmt "... wrote `~A'\n" (fix-path scm-file)))
       (lambda (key fmt . args)
+	(if (access? scm-file W_OK) (delete-file scm-file))
 	(apply fail fmt args)
 	(exit 1)))
     (unless (assq-ref options 'no-exec)
-      (sfmt "compiling `~A' ...\n" (cleanup scm-file))
-       (let ((go-file (compile-file scm-file
+      (sfmt "compiling `~A' ...\n" (fix-path scm-file))
+      (let ((go-file (compile-file scm-file
 				   #:from 'scheme #:to 'bytecode
 				   #:opts '())))
 	(load-compiled go-file)
-	(sfmt "... wrote `~A'\n" (cleanup go-file)))
+	(sfmt "... wrote `~A'\n" (basename go-file)))
       (sleep 1))))
 
 (define (main . args)
