@@ -29,6 +29,7 @@
 	    ;; maybe used outside of modules?
 	    fh:cast fh-cast bs-addr
 	    fh:function
+	    fh:pointer
 	    ffi-void*
 	    make-fht
 
@@ -37,9 +38,8 @@
 	    define-fh-type-alias
 	    define-fh-compound-type
 	    define-fh-vector-type
-	    define-fh-vector&-type
 	    define-fh-function*-type
-	    ref<->deref! fh-ref<->deref!
+	    fh-ref<=>deref! ref<->deref!
 	    make-symtab-function
 	    fh-find-symbol-addr
 	    fht-wrap fht-unwrap fh-wrap fh-unwrap
@@ -47,6 +47,9 @@
 	    unwrap~pointer unwrap~array
 	    make-fctn-param-unwrapper
 	    fh-link-proc fh-link-extern
+
+	    ;; commonly used libc functions
+	    fopen fclose
 	    
 	    ;; deprecated
 	    fh-link-bstr ;; => fh-link-extern
@@ -56,9 +59,9 @@
   #:use-module (rnrs bytevectors)
   #:use-module ((system foreign) #:prefix ffi:)
   #:use-module (srfi srfi-9)
-  #:version (1 01 2))
+  #:version (1 02 0))
 
-(define *ffi-help-version* "1.01.2")
+(define *ffi-help-version* "1.02.0")
 
 (define (sferr fmt . args) (apply simple-format (current-error-port) fmt args))
 
@@ -84,18 +87,15 @@
 ;; We should check with struct-vtable?
 ;; name as symbol
 (define* (make-fht name unwrap wrap pointer-to value-at printer)
-  ;;(simple-format #t "make-fht: ~S\n" name)
   (let* ((ty (make-struct/no-tail
 	      ffi-helper-type
 	      (make-struct-layout "pw") ;; 1 slot for value
 	      printer
 	      (or unwrap (lambda (obj) (error "no unwrapper")))
 	      (or wrap (lambda (obj) (error "no wrapper")))
-	      ;;(or pointer-to (lambda (obj) (error "no pointer-to")))
 	      (or pointer-to (lambda (obj) (ffi:bytevector->pointer
 					    (bytestructure-bytevector
 					     (fh-object-val obj)))))
-	      ;;(or value-at (lambda (obj) (error "no value-at")))
 	      (or value-at (lambda (obj) (bytestructure-ref
 					  (fh-object-val obj) '*)))))
 	 (vt (struct-vtable ty)))
@@ -280,18 +280,18 @@
 	  (else (make-struct/no-tail type val))))
 	(() (make 0))))))
 
-;; @deffn {Syntax} ref<->deref! p-type p-make type make
+;; @deffn {Syntax} fh-ref<=>deref! p-type p-make type make
 ;; This procedure will ``connect'' the two types so that the procedures
 ;; @code{pointer-to} and @code{value-at} work.
 ;; @end deffn
-(define (ref<->deref! p-type p-make type make)
+(define (fh-ref<=>deref! p-type p-make type make)
   (if p-make
       (struct-set! type (+ vtable-offset-user 2) ; pointer-to
 		   (lambda (obj) (p-make (bs-addr (fh-object-val obj))))))
   (if make
       (struct-set! p-type (+ vtable-offset-user 3) ; value-at
 		   (lambda (obj) (make (fh-object-ref obj '*))))))
-(define fh-ref<->deref! ref<->deref!)
+(define ref<->deref! fh-ref<=>deref!)
 
 ;; @deffn {Syntax} define-fh-type-alias alias type
 ;; set up type alias.  Caller needs to match type? and make
@@ -352,52 +352,91 @@
 
 ;; == extension to bytestructures ==============================================
 
-;; unwrapper
-;; setter
-;; getter
+;; adopted from code at  https://github.com/TaylanUB covered by GPL3+ and
+;; Copyright (C) 2015 Taylan Ulrich BayirliKammer <taylanbayirli@gmail.com>
 
-;; variable length vector bytestructure -- double v[];
-;; calling this vector&; has zero size
-;;   (define d (fh:vector& double))
-;;   (define b (make-bytevector 80))
-;;   (define bs (make-bytestructure b 0 d))
-;;   (bytestructure-ref bs 0) => 0.0
-(define-record-type <vector&-metadata>
-  (make-vector&-metadata element-descriptor)
-  vector&-metadata?
-  (element-descriptor vector&-metadata-element-descriptor))
+(define make-pointer-metadata
+  (@@ (bytestructures guile pointer) make-pointer-metadata))
 
-(define (fh:vector& descriptor)
-  (define element-size (bytestructure-descriptor-size descriptor))
-  (define size 0)
-  (define alignment (bytestructure-descriptor-alignment descriptor))
+(define (fh:pointer %descriptor)
+  (define pointer-size (ffi:sizeof '*))
+  (define bytevector-address-ref
+    (case pointer-size
+      ((1) bytevector-u8-ref)
+      ((2) bytevector-u16-native-ref)
+      ((4) bytevector-u32-native-ref)
+      ((8) bytevector-u64-native-ref)))
+  (define bytevector-address-set!
+    (case pointer-size
+      ((1) bytevector-u8-set!)
+      ((2) bytevector-u16-native-set!)
+      ((4) bytevector-u32-native-set!)
+      ((8) bytevector-u64-native-set!)))
+  (define (pointer-ref bytevector offset content-size)
+    (let ((address (bytevector-address-ref bytevector offset)))
+      (if (zero? address)
+	  (error "Tried to dereference null-pointer.")
+	  (ffi:pointer->bytevector (ffi:make-pointer address) content-size))))
+  (define (pointer-idx-ref bytevector offset index content-size)
+    (let* ((base-address (bytevector-address-ref bytevector offset))
+	   (address (+ base-address (* index content-size))))
+      (if (zero? base-address)
+	  (error "Tried to dereference null-pointer.")
+	  (ffi:pointer->bytevector (ffi:make-pointer address) content-size))))
+  (define (pointer-set! bytevector offset value)
+    (cond
+     ((exact-integer? value)
+      (bytevector-address-set! bytevector offset value))
+     ((bytevector? value)
+      (bytevector-address-set! bytevector offset
+			       (ffi:bytevector->pointer value)))
+     ((bytestructure? value)
+      (bytevector-address-set! bytevector offset
+			       (ffi:bytevector->pointer
+				(bytestructure-bytevector value))))))
+  (define (get-descriptor)
+    (if (promise? %descriptor)
+        (force %descriptor)
+        %descriptor))
+  (define size pointer-size)
+  (define alignment size)
   (define (unwrapper syntax? bytevector offset index)
-    (if syntax? (throw 'ffi-help-error "fh:vector& syntax not supported"))
-    (values bytevector (+ offset (* index element-size)) descriptor))
-  (define meta (make-vector&-metadata descriptor))
-  (make-bytestructure-descriptor size alignment unwrapper #f #f meta))
-(export fh:vector&)
-
-(define-syntax-rule (define-fh-vector&-type type base-desc type? make)
-  (begin
-    (define type
-      (make-fht
-       (quote type)
-       (lambda (obj) ;; unwrap => pointer
-	 (ffi:bytevector->pointer (bytestructure-bytevector obj)))
-       (lambda* (val #:optional (size 0))
-	 (cond
-	  ((bytevector? val)
-	   (make-struct/no-tail
-	    type (make-bytestructure val 0 (fh:vector& base-desc))))
-	  ((ffi:pointer? val) (make (ffi:pointer->bytevector val size)))
-	  ((integer? val) (make (ffi:make-pointer val) size))
-	  (else (error "ffi-help-rt: bad spec" val))))
-       #f #f
-       (make-bs-printer (quote type))))
-    (define (type? obj)
-      (and (fh-object? obj) (eq? (struct-vtable obj) type)))
-    (define make (fht-wrap type))))
+    (define (syntax-list id . elements)
+      (datum->syntax id (map syntax->datum elements)))
+    (let ((descriptor (get-descriptor)))
+      (when (eq? 'void descriptor)
+        (error "Tried to follow void pointer."))
+      (let* ((size (bytestructure-descriptor-size descriptor))
+             (index-datum (if syntax? (syntax->datum index) index)))
+        (cond
+	 ((eq? '* index-datum)
+	  (if syntax?
+	      (values #`(pointer-ref #,bytevector #,offset #,size) 0 descriptor)
+	      (values (pointer-ref bytevector offset size) 0 descriptor)))
+	 ((integer? index-datum)
+	  (if syntax?
+	      (values #`(pointer-idx-ref #,bytevector #,offset ,index #,size)
+		      0 descriptor)
+	      (values (pointer-idx-ref bytevector offset index-datum size)
+		      0 descriptor)))
+	 (else
+	  (if syntax?
+	      (let ((bytevector* #`(pointer-ref #,bytevector #,offset #,size)))
+                (bytestructure-unwrap/syntax
+                 bytevector* 0 descriptor (syntax-list index index)))
+	      (let ((bytevector* (pointer-ref bytevector offset size)))
+                (bytestructure-unwrap*
+                 bytevector* 0 descriptor index))))))))
+  (define (getter syntax? bytevector offset)
+    (if syntax?
+        #`(bytevector-address-ref #,bytevector #,offset)
+        (bytevector-address-ref bytevector offset)))
+  (define (setter syntax? bytevector offset value)
+    (if syntax?
+        #`(pointer-set! #,bytevector #,offset #,value)
+        (pointer-set! bytevector offset value)))
+  (define meta (make-pointer-metadata %descriptor))
+  (make-bytestructure-descriptor size alignment unwrapper getter setter meta))
 
 ;; @deffn {Procedure} fh:function return-desc param-desc-list
 ;; @deffnx {Syntax} define-fh-function*-type name desc type? make
@@ -413,6 +452,9 @@
   (return-descriptor function-metadata-return-descriptor)
   (param-descriptor-list function-metadata-param-descriptor-list)
   (attributes function-metadata-attributes))
+(export function-metadata?
+	function-metadata-return-descriptor
+	function-metadata-param-descriptor-list)
 
 (define (pointer->procedure/varargs return-ffi pointer param-ffi-list)
   (define (arg->ffi arg)
@@ -498,8 +540,8 @@
    ((eq? bs-desc 'void) ffi:void)
    (else (error "missed type"))))
 
-;; given a fs:function return pair: (return-type . arg-list)
-(define (fs-function*-signature desc)
+;; given a fh:function return pair: (return-type . arg-list)
+(define (fh-function*-signature desc)
   (let* ((meta (bytestructure-descriptor-metadata desc))
 	 (desc (pointer-metadata-content-descriptor meta))
 	 (desc (if (promise? desc) (force desc) desc))
@@ -509,6 +551,7 @@
 	 (bs-al (function-metadata-param-descriptor-list meta))
 	 (ffi-bs-al (map bs-desc->ffi-desc bs-al)))
     (cons ffi-rt ffi-bs-al)))
+(define fs-function*-signature fh-function*-signature)
 
 ;; @deffn {Syntax} define-fh-function*-type type desc type? make
 ;; document this
@@ -524,7 +567,7 @@
 	  (lambda (obj)
 	    (cond
 	     ((procedure? obj)		; a lambda
-	      (let* ((sig (fs-function*-signature desc)))
+	      (let* ((sig (fh-function*-signature desc)))
 		(ffi:procedure->pointer (car sig) obj (cdr sig))))
 	     ((and (pair? obj) (fh-type? (car obj))) ; fh-cast
 	      (unwrap~pointer (cdr obj)))
@@ -546,7 +589,7 @@
 	     ((number? val) (bytestructure desc val))
 	     ((ffi:pointer? val) (bytestructure desc (ffi:pointer-address val)))
 	     ((procedure? val) ;; special case, proceadure not pointer
-	      (let* ((sig (fs-function*-signature desc)))
+	      (let* ((sig (fh-function*-signature desc)))
 		(bytestructure
 		 desc
 		 (ffi:pointer-address
@@ -657,7 +700,7 @@
 (define-fh-pointer-type char** char**-desc char**? make-char**)
 (export char**-desc char** char**? make-char**)
 
-(ref<->deref! char** make-char** char* make-char*)
+(fh-ref<=>deref! char** make-char** char* make-char*)
 
 (define (char*->string obj)
   (ffi:pointer->string (ffi:make-pointer (fh-object-ref obj))))
@@ -701,7 +744,7 @@
 	      (display ">" port))))
 (define make-void* (fht-wrap void*))
 (define void*? (lambda (obj) (eq? (struct-vtable obj) void*)))
-(ref<->deref! void* make-void* fh-void make-fh-void)
+(fh-ref<=>deref! void* make-void* fh-void make-fh-void)
 (export void* void*? make-void*)
 
 (define void**-desc (bs:pointer (bs:pointer 'void)))
@@ -720,7 +763,7 @@
 	      (display ">" port))))
 (define make-void** (fht-wrap void**))
 (define void**? (lambda (obj) (eq? (struct-vtable obj) void**)))
-(ref<->deref! void** make-void** void* make-void*)
+(fh-ref<=>deref! void** make-void** void* make-void*)
 (export void** void**? make-void**)
 
 
@@ -820,6 +863,28 @@
 	 (size (bytestructure-descriptor-size desc)))
     (make-bytestructure (ffi:pointer->bytevector addr size) 0 desc)))
 
+
+;; === common c functions called
+
+;; @deffn {Procedure} fopen filename mode
+;; Call the C fucntion fopen and return a scheme @code{<pointer>} type.
+;; @end deffn
+(define fopen
+  (let ((~fopen (ffi:pointer->procedure
+		 '* (dynamic-func "fopen" (dynamic-link)) (list '* '*))))
+    (lambda (filename mode)
+      (~fopen (ffi:string->pointer filename) (ffi:string->pointer mode)))))
+
+;; @deffn {Procedure} fopen file
+;; Call the C fucntion fclose on @var<file>, a @code{<pointer>} type generated
+;; by @code{fopen}.
+;; @end deffn
+(define fclose
+  (let ((~fclose (ffi:pointer->procedure
+		 ffi:int (dynamic-func "fclose" (dynamic-link)) (list '*))))
+    (lambda (file)
+      (~fclose file))))
+
 ;; === deprecated ==============================================================
 
 (define fh-link-bstr fh-link-extern)
